from datetime import datetime
from traceback import format_exception


class FileHandler(object):
    def __init__(self, path='var/log/fail.log', mode='a'):
        self.logfile = open(path, mode)

    def handle_error(self, exc_info, ctx=None):
        if ctx:
            ctx_data = ''.join(
                '%s: %s\n' % (x, y) for (x, y)
                in sorted(ctx.items(), key=lambda x: x[0])
            )
        else:
            ctx_data = ''
        self.logfile.write(
            '%s'
            '%s\n'
            '---\n' % (
                ctx_data,
                ''.join(format_exception(*exc_info)),
            )
        )
        self.logfile.flush()


class MongodbHandler(object):
    def __init__(self, **kwargs):
        from pymongo import MongoClient

        self.db_name = kwargs.pop('database', 'ioweb')
        self.col_name = kwargs.pop('collection', 'crawler_error')
        self.connection_args = kwargs
        self.db = MongoClient(**kwargs)[self.db_name]

    def handle_error(self, exc_info, ctx=None):
        ctx = ctx or {}
        ctx['traceback'] = ''.join(format_exception(*exc_info))
        ctx['error_type'] = exc_info[0].__name__
        ctx['error_msg'] = str(exc_info[1])
        self.db[self.col_name].insert_one({
            'date': datetime.utcnow(),
            'data': ctx,
        })


class ErrorLogger(object):
    aliases = {
        'file': FileHandler,
    }

    def __init__(self):
        self.handlers = []

    def add_handler(self, hdl, remove_handlers=False):
        if remove_handlers:
            self.handlers = []
        if isinstance(hdl, str):
            hdl = self.aliases[hdl]() 
        self.handlers.append(hdl)

    def log_error(self, exc_info, ctx=None):
        for hdl in self.handlers:
            hdl.handle_error(exc_info, ctx)



from pprint import pprint

from .response import Response
from .request import Request
from .transport import Urllib3Transport
from .error import NetworkError


class Session(object):
    def __init__(self):
        self.req = Request()
        self.transport = Urllib3Transport()

    def setup(self, **kwargs):
        self.req.setup(**kwargs)

    def request(self, url=None, **kwargs):
        self.setup(url=url, **kwargs)
        res = Response()
        self.transport.prepare_request(self.req, res)
        self.transport.request(self.req, res)
        self.transport.prepare_response(self.req, res, None)
        return res



from pprint import pprint
import logging
import time
from contextlib import contextmanager
import traceback
import sys
from urllib.parse import urlencode

from urllib3.filepost import encode_multipart_formdata
from urllib3.util.retry import Retry
from urllib3.util.timeout import Timeout
from urllib3 import exceptions, ProxyManager, make_headers
from  urllib3.contrib import pyopenssl
from urllib3.contrib.socks import SOCKSProxyManager
import urllib3
import OpenSSL.SSL
import certifi

from . import error
from .urllib3_custom import CustomPoolManager

urllib3.disable_warnings(exceptions.InsecureRequestWarning)
pyopenssl.inject_into_urllib3()


class Urllib3Transport(object):
    __slots__ = (
        'urllib3_response',
        'op_started',
        'pools',
    )

    def __init__(self, pools=None):
        if pools is None:
            pools = {}
        self.pools = {}
        self.pools[(None, None, True)] = CustomPoolManager(
            cert_reqs='CERT_REQUIRED',
            ca_certs=certifi.where(),
        )
        self.pools[(None, None, False)] = CustomPoolManager()
        self.urllib3_response = None

    def prepare_request(self, req, res):
        pass

    @contextmanager
    def handle_network_error(self, req):
        try:
            yield
        except exceptions.ReadTimeoutError as ex:
            raise error.OperationTimeoutError(str(ex), ex)
        except exceptions.ConnectTimeoutError as ex:
            raise error.ConnectError(str(ex), ex)
        except exceptions.ProtocolError as ex:
            raise error.ConnectError(str(ex), ex)
        except exceptions.SSLError as ex:
            raise error.ConnectError(str(ex), ex)
        except OpenSSL.SSL.Error as ex:
            raise error.ConnectError(str(ex), ex)
        except exceptions.LocationParseError as ex:
            raise error.MalformedResponseError(str(ex), ex)
        except exceptions.LocationValueError as ex:
            raise error.InvalidUrlError(str(ex), ex)
        except exceptions.DecodeError as ex:
            raise error.MalformedResponseError(str(ex), ex)
        except exceptions.InvalidHeader as ex:
            raise error.MalformedResponseError(str(ex), ex)
        except exceptions.ProxyError as ex:
            raise error.ProxyError(str(ex), ex)
        except exceptions.MaxRetryError as ex:
            raise error.TooManyRedirects(str(ex), ex)
        except AttributeError:
            # See https://github.com/urllib3/urllib3/issues/1556
            etype, evalue, tb = sys.exc_info()
            frames = traceback.extract_tb(tb)
            found = False
            for frame in frames:
                if (
                        "if host.startswith('[')" in frame.line
                        and 'connectionpool.py' in frame.filename
                    ):
                    found = True
                    break
            if found:
                raise error.MalformedResponseError('Invalid redirect header')
            else:
                raise
        except UnicodeEncodeError as ex:
            # UnicodeEncodeError is subclass of ValueError
            etype, evalue, tb = sys.exc_info()
            frames = traceback.extract_tb(tb)
            #    self._output(request.encode('ascii'))
            #UnicodeEncodeError: 'ascii' codec can't encode characters
            #in position 15-24: ordinal not in range(128)
            if 'self._output(request.encode' in frames[-1].line:
                raise error.InvalidUrlError(
                    'Non ASCII URL: %s' % req.config.get('url')
                )
            else:
                raise
        except ValueError as ex:
            if 'Invalid IPv6 URL' in str(ex):
                raise error.MalformedResponseError('Invalid redirect header')
            else:
                raise
        except KeyError as ex:
            etype, evalue, tb = sys.exc_info()
            frames = traceback.extract_tb(tb)
            if 'self.key_fn_by_scheme[scheme]' in frames[-1].line:
                raise error.InvalidUrlError(
                    'Unknown URL scheme: %s' % req.config.get('url')
                )
            else:
                raise



    def get_pool(self, req):
        if req['proxy']:
            if req['proxy_auth']:
                proxy_headers = make_headers(proxy_basic_auth=req['proxy_auth'])
            else:
                proxy_headers = None
            proxy_url = '%s://%s' % (req['proxy_type'], req['proxy'])
            pool_key = (req['proxy_type'], req['proxy'], bool(req['verify']))
            if pool_key not in self.pools:
                if req['proxy_type'] == 'socks5':
                    opts = {
                        #num_pools=1000,
                        #maxsize=10,
                    }
                    if req['verify']:
                        pool = SOCKSProxyManager(
                            proxy_url,
                            cert_reqs='CERT_REQUIRED',
                            ca_certs=certifi.where(),
                            **opts
                        )
                    else:
                        pool = SOCKSProxyManager(proxy_url, **opts)
                elif req['proxy_type'] == 'http':
                    opts = {
                        #num_pools=1000,
                        #maxsize=10,
                    }
                    if req['verify']:
                        pool = ProxyManager(
                            proxy_url,
                            proxy_headers=proxy_headers,
                            cert_reqs='CERT_REQUIRED',
                            ca_certs=certifi.where(),
                            **opts,
                        )
                    else:
                        pool = ProxyManager(
                            proxy_url,
                            proxy_headers=proxy_headers,
                            **opts
                        )
                else:
                    raise IowebConfigError(
                        'Invalid value of request option `proxy_type`: %s'
                        % req['proxy_type']
                    )
                self.pools[pool_key] = pool
            else:
                pool = self.pools[pool_key]
        else:
            pool = self.pools[(None, None, bool(req['verify']))]
        return pool

    def request(self, req, res):
        options = {}
        headers = req['headers'] or {}

        #req_url = req['url']
        #if not '://' in req_url:
        #    req_url = 'http://%s' % req_url
        #url_parts = urlsplit(req_url)
        #if url_parts.scheme not in ('http', 'https'):
        #    emsg = 'Unknown scheme [%s]' % url_parts.scheme
        #    # second argument saved as `err.transport_error`
        #    # and used then to generate short code of error
        #    # for statistics
        #    raise error.InvalidUrlError(
        #        emsg, error.InvalidUrlError(emsg),
        #    )

        pool = self.get_pool(req)

        self.op_started = time.time()
        if req['resolve']:
            if req['proxy']:
                raise error.IowebConfigError(
                    'Request option `resolve` could not be used along option `proxy`'
                )
            for host, ip in req['resolve'].items():
                pool.resolving_cache[host] = ip


        if req['content_encoding']:
            if not any(x.lower() == 'accept-encoding' for x in headers):
                headers['Accept-Encoding'] = req['content_encoding']

        if req['data']:
            if isinstance(req['data'], dict):
                if req['multipart']:
                    body, ctype = encode_multipart_formata(req['data'])
                else:
                    body = urlencode(req['data'])
                    ctype = 'application/x-www-form-urlencoded'
                options['body'] = body
                headers['Content-Type'] = ctype
            elif isinstance(req['data'], bytes):
                options['body'] = req['data']
            elif isinstance(req['data'], str):
                options['body'] = req['data'].encode('utf-8')
            else:
                raise IowebConfigError(
                    'Invalid type of request data option: %s'
                    % type(req['data'])
                )
            headers['Content-Length'] = len(options['body'])

        with self.handle_network_error(req):
            if req['follow_redirect']:
                retry_opts = {
                    'redirect': req['max_redirects'],
                    'raise_on_redirect': True,
                }
            else:
                retry_opts = {
                    'redirect': False,
                    'raise_on_redirect': False,
                }
            #try:
            self.urllib3_response = pool.urlopen(
                req.method(),
                req['url'],
                headers=headers,
                # total - set to None to remove this constraint
                # and fall back on other counts. 
                retries=Retry(
                    total=None,
                    connect=False,
                    read=False,
                    **retry_opts,
                ),
                timeout=Timeout(
                    connect=req['connect_timeout'],
                    read=req['timeout'],
                ),
                preload_content=False,
                decode_content=req['decode_content'],
                **options
            )
            #except UnicodeEncodeError as ex:
            #    print('ERROR', ex)
            #    print('URL', req['url'])
            #    import pdb; pdb.set_trace()
            #    #raise

    def read_with_timeout(self, req, res):
        read_limit = req['content_read_limit']
        chunk_size = 1024
        bytes_read = 0
        while True:
            chunk = self.urllib3_response.read(chunk_size)
            if chunk:
                if read_limit:
                    chunk_limit = min(len(chunk), read_limit - bytes_read)
                else:
                    chunk_limit = len(chunk)
                res._bytes_body.write(chunk[:chunk_limit])
                bytes_read += chunk_limit
                if read_limit and bytes_read >= read_limit:
                    break
            else:
                break
            if time.time() - self.op_started > req['timeout']:
                raise error.OperationTimeoutError(
                    'Timed out while reading response',
                )

    def prepare_response(self, req, res, err, raise_network_error=True):
        try:
            if err:
                res.error = err
            else:
                try:
                    res.headers = self.urllib3_response.headers
                    res.status = self.urllib3_response.status

                    if hasattr(self.urllib3_response._connection.sock, 'connection'):
                        res.cert = (
                            self.urllib3_response._connection.sock.connection
                            .get_peer_cert_chain()
                        )

                    with self.handle_network_error(req):
                        self.read_with_timeout(req, res)
                except error.NetworkError as ex:
                    if raise_network_error:
                        raise
                    else:
                        res.error = err
        finally:
            if self.urllib3_response:
                self.urllib3_response.release_conn()



class BaseRequest(object):
    __slots__ = (
        'config',
        'retry_count',
        'priority',
        'meta',
        'retry_errors',
        'error_context',
    )
    valid_config_keys = () 

    def __init__(
            self, retry_count=0, priority=100, meta=None,
            retry_errors=None, error_context=None,
            **kwargs
        ):
        self.retry_count = retry_count
        self.priority = priority
        self.meta = meta or {}
        self.config = self.get_default_config()
        assert \
            retry_errors is None or isinstance(retry_errors, tuple), \
            'retry_errors must be None or tuple'
        self.retry_errors = retry_errors
        self.error_context = error_context
        self.setup(**kwargs)


    def get_default_config(self):
        return {}

    def setup(self, **kwargs):
        for key in kwargs:
            assert key in self.valid_config_keys, 'Invalid configuration key: %s' % key
            self.config[key] = kwargs[key]

    def __getitem__(self, key):
        return self.config[key]

    def as_data(self):
        return {
            'config': self.config,
            'retry_count': self.retry_count,
            'priority': self.priority,
            'meta': self.meta,
        }

    @classmethod
    def from_data(cls, data):
        # TODO: check config keys against `get_default_config` keys
        req = Request()
        for key in ('config', 'retry_count', 'priority', 'meta'):
            setattr(req, key, data[key])
        return req

    def __lt__(self, other):
        return self.priority < other.priority

    def __eq__(self, other):
        #if not self.priority or not other.priority:
        #    return True
        #else:
        return self.priority == other.priority


class Request(BaseRequest):
    # TODO: remove, use keys from default_config_keys
    valid_config_keys = (
        'name', 'url',
        'max_redirects',
        'follow_redirect',
        'timeout',
        'connect_timeout',
        'resolve',
        'raw',
        'headers',
        'content_encoding',
        'decode_content',
        'content_read_limit',
        'priority',
        'extra_valid_status',
        'proxy',
        'proxy_auth',
        'proxy_type',
        'data',
        'multipart',
        'verify',
        'method',
    )

    def get_default_config(self):
        return {
            'method': None,
            'name': None,
            'url': None,
            'max_redirects': 5,
            'follow_redirect': True,
            'timeout': 10,
            'connect_timeout': 5,
            'resolve': None,
            'raw': False,
            'headers': None,
            'content_encoding': 'gzip,chunked', 
            'decode_content': True,
            'content_read_limit': None,
            'priority': 100,
            'extra_valid_status': None,
            'proxy': None,
            'proxy_auth': None,
            'proxy_type': 'http',
            'data': None,
            'multipart': False,
            'verify': True,
        }

    def method(self):
        if self.config['method']:
            return self.config['method']
        elif self.config['data']:
            return 'POST'
        else:
            return 'GET'


class CallbackRequest(BaseRequest):
    """
    CallbackRequest is not processed with network transport.
    It has `network_callback` which is
    executed instead of `transport.request`
    """
    # TODO: remove, use keys from default_config_keys
    valid_config_keys = (
        'name',
        'timeout',
        'raw',
        'network_callback',
    )

    def get_default_config(self):
        return {
            'name': None,
            'timeout': 10,
            'raw': False,
            'network_callback': None,
        }



from queue import Empty
import time
import sys
import logging
from collections import deque
from threading import Thread
from uuid import uuid4
import traceback

from urllib3 import PoolManager
import gevent
from gevent import Timeout

from .transport import Urllib3Transport
from .util import debug
from .response import Response
from .error import (
    DataNotValid, NetworkError, OperationTimeoutError, collect_error_context
)
from .request import Request, CallbackRequest

network_logger = logging.getLogger(__name__)


class NetworkService(object):
    def __init__(
            self,
            taskq,
            resultq,
            fatalq,
            resultq_size_limit=None,
            threads=3,
            shutdown_event=None,
            pause=None,
            setup_request_hook=None,
            setup_request_proxy=None,
            stat=None,
        ):
        # Input arguments
        self.taskq = taskq
        self.resultq = resultq
        self.fatalq = fatalq
        if resultq_size_limit is None:
            resultq_size_limit = threads * 2
        self.resultq_size_limit = resultq_size_limit
        self.threads = threads
        self.shutdown_event = shutdown_event
        self.pause = pause
        self.setup_request_hook = setup_request_hook
        self.setup_request_proxy = setup_request_proxy
        self.stat = stat
        # Init logic
        self.idle_handlers = set()
        self.active_handlers = set()
        self.registry = {}
        self.pools = {}
        for _ in range(threads):
            ref = object()
            self.idle_handlers.add(ref)
            self.registry[ref] = {
                'transport': Urllib3Transport(
                    #pools=self.pools,
                ),
                'request': None,
                'response': None,
                'start': None,
            }

    def run(self):
        task = None

        while not self.shutdown_event.is_set():
            if self.pause.pause_event.is_set():
                if (
                        task is None
                        and not len(self.active_handlers)
                        and len(self.idle_handlers) == self.threads
                    ):
                    self.pause.process_pause()

            if (
                    task is None
                    and
                    self.resultq.qsize() < self.resultq_size_limit
                ):
                try:
                    prio, task = self.taskq.get(False)
                except Empty:
                    pass

            # TODO: convert idle_handlers into queue, blocking wait on
            # next idle handler if task available
            if task:
                if len(self.idle_handlers):
                    self.start_request_thread(task)
                    task = None
                else:
                    time.sleep(0.01)
            else:
                time.sleep(0.01)

    def start_request_thread(self, req):
        ref = self.idle_handlers.pop()
        transport = self.registry[ref]['transport']
        res = Response()
        transport.prepare_request(req, res)
        self.active_handlers.add(ref)
        self.registry[ref].update({
            'request': req,
            'response': res,
            'start': time.time(),
        })

        if self.setup_request_hook:
            self.setup_request_hook(transport, req)

        if self.setup_request_proxy:
            self.setup_request_proxy(transport, req)

        gevent.spawn(
            self.thread_network,
            ref,
            transport,
            req,
            res
        )

    def log_network_request(self, req):
        if isinstance(req, Request):
            if req.retry_count > 0:
                retry_str = ' [retry=#%d]' % req.retry_count
            else:
                retry_str = ''
            if req['proxy']:
                proxy_str = ' [proxy=%s, type=%s, auth=%s]' % (
                    req['proxy'],
                    req['proxy_type'].upper(),
                    ('YES' if req['proxy_auth'] else 'NO'),
                )
            else:
                proxy_str = ''
            network_logger.debug(
                '%s %s%s%s', req.method(), req['url'], retry_str, proxy_str
            )

    def thread_network(self, ref, transport, req, res):
        try:
            self.log_network_request(req)
            try:
                timeout_time = req['timeout'] or 31536000
                with Timeout(
                        timeout_time,
                        OperationTimeoutError(
                            'Timed out while reading response',
                            Timeout(timeout_time),
                        )
                    ):
                    if isinstance(req, CallbackRequest):
                        req['network_callback'](req, res)
                    else:
                        transport.request(req, res)
            except OperationTimeoutError as ex:
                #logging.error(ex)
                error = ex
            except (req.retry_errors or (NetworkError, DataNotValid)) as ex:
                #logging.error(ex)
                error = ex
            except Exception as ex:
                #logging.error(ex)
                raise
            else:
                error = None
            if isinstance(req, CallbackRequest):
                res.error = error
            else:
                transport.prepare_response(
                    req, res, error, raise_network_error=False
                )
            self.resultq.put({
                'request': req,
                'response': res,
            })
        except Exception as ex:
            ctx = collect_error_context(req)
            self.fatalq.put((sys.exc_info(), ctx))
        finally:
            self.free_handler(ref)

    def free_handler(self, ref):
        self.active_handlers.remove(ref)
        self.idle_handlers.add(ref)
        self.registry[ref]['request'] = None
        self.registry[ref]['response'] = None
        self.registry[ref]['start'] = None



import pytest

from ioweb.session import Session
from ioweb.error import IowebConfigError

from .fixtures import test_server, TEST_SERVER


def test_resolve(test_server):
    sess = Session()
    res = sess.request(
        url='http://ya.ru:%d/' % TEST_SERVER['port'],
        resolve={
            'ya.ru': '127.0.0.1',
        }
    )
    assert res.text == TEST_SERVER['content']


#def test_resolve_proxy(test_server):
#    sess = Session()
#    with pytest.raises(IowebConfigError):
#        res = sess.request(
#            url='http://ya.ru:%d/' % TEST_SERVER['port'],
#            resolve={
#                'ya.ru': '127.0.0.1',
#            },
#            proxy='1.1.1.1',
#        )



import time
import sys
import logging
from threading import Event, Thread, Lock
from queue import Queue, PriorityQueue, Empty, Full
from urllib.parse import urlsplit 
import re
import json
from urllib.request import urlopen
from traceback import format_exception
from collections import defaultdict
import gc
from copy import deepcopy
from datetime import datetime

from .util import Pause, debug
#from .loop import MultiCurlLoop
from .network_service import NetworkService
from .stat import Stat
from .request import Request, CallbackRequest, BaseRequest
from .error import get_error_tag, collect_error_context
from .error_logger import ErrorLogger
from .proxylist import ProxyList


class Crawler(object):
    _taskgen_sleep_time = 0.01
    dataop_threshold_default = {
        'number': 500,
        'size': None,
        #'time': None,
    }
    dataop_threshold = {}

    def task_generator(self):
        if False:
            yield None

    def __init__(self,
            network_threads=3,
            result_workers=4,
            retry_limit=3,
            extra_data=None,
            stop_on_handler_error=False,
            debug=False,
        ):
        if extra_data is None:
            self.extra_data = {}
        else:
            self.extra_data = deepcopy(extra_data)
        self.taskq = PriorityQueue()
        self.taskq_size_limit = max(100, network_threads * 2)
        self.resultq = Queue()
        self.shutdown_event = Event()
        self.fatal_error_happened = Event()
        self.network_pause = Pause()
        self.retry_limit = retry_limit
        self.stat = Stat(speed_keys='crawler:request-processed')
        self.fatalq = Queue()
        self.network = NetworkService(
            self.taskq, self.resultq,
            fatalq=self.fatalq,
            threads=network_threads,
            shutdown_event=self.shutdown_event,
            pause=self.network_pause,
            setup_request_hook=self.setup_request_hook,
            setup_request_proxy=self.setup_request_proxy,
            stat=self.stat,
        )
        self.result_workers = result_workers
        self._run_started = None

        self.dataopq = {}
        self.dataop_lock = defaultdict(Lock)
        self.dataop_counters = defaultdict(lambda: {
            'number': 0,
            'size': 0,
        })
        self.error_logger = ErrorLogger()
        self.stop_on_handler_error = stop_on_handler_error
        self.proxylist = None
        self.debug = debug

        self.init_hook()

    @classmethod
    def update_arg_parser(cls, parser):
        for key, config in cls.extra_cli_args().items():
            parser.add_argument('--extra-%s' % key, **config)

    @classmethod
    def extra_cli_args(cls):
        return {}

    def load_proxylist(self, pl_type, pl_location, **kwargs):
        self.proxylist = ProxyList.create_from_source(
            pl_type, pl_location, **kwargs
        )

    def is_dataopq_dump_time(self, name):
        to_dump = False
        try:
            number_th = self.dataop_threshold[name]['number']
        except KeyError:
            number_th = self.dataop_threshold_default['number']
        if number_th:
            if self.dataop_counters[name]['number'] >= number_th:
                    to_dump = True
        if not to_dump:
            try:
                size_th = self.dataop_threshold[name]['size']
            except KeyError:
                size_th = self.dataop_threshold_default['size']
            if size_th:
                if self.dataop_counters[name]['size'] >= size_th:
                    to_dump = True
        return to_dump

    def enq_dataop(self, name, op, size=None, force_dump=False):
        if op:
            self.enq_dataop_hook(name, op)
        self.dataop_lock[name].acquire()
        released = False
        try:
            self.dataopq.setdefault(name, [])
            if op:
                self.dataopq[name].append(op)
                self.dataop_counters[name]['number'] += 1
                if size:
                    self.dataop_counters[name]['size'] += size

            if force_dump or self.is_dataopq_dump_time(name):
                #logging.debug('Dumping data ops')
                self.stat.inc('dataop-dump-%s' % name)
                ops = self.dataopq[name]
                if ops:
                    self.dataopq[name] = []
                    self.dataop_counters[name]['number'] = 0
                    self.dataop_counters[name]['size'] = 0
                    # release as soon as possible
                    self.dataop_lock[name].release()
                    released = True
                    func = getattr(self, 'dataop_handler_%s' % name)
                    func(ops)
                else:
                    self.dataop_lock[name].release()
                    released = True
        finally:
            if not released:
                self.dataop_lock[name].release()

    def enq_dataop_hook(self, name, op):
        pass

    def init_hook(self):
        pass

    def setup_request_proxy(self, transport, req):
        if self.proxylist:
            proxy = self.proxylist.random_server()
            req.setup(
                proxy=proxy.address(),
                proxy_auth=proxy.auth(),
                proxy_type=proxy.proxy_type,
            )

    def setup_request_hook(self, transport, req):
        pass

    def submit_task(self, req):
        self.submit_task_hook(req)
        self.taskq.put((req.priority, req))

    def submit_task_hook(self, req):
        pass

    def thread_task_generator(self):
        try:
            try:
                tgen_iter = iter(self.task_generator())
            except TypeError:
                return
            else:
                for item in tgen_iter:
                    while item:
                        if self.shutdown_event.is_set():
                            return
                        if self.taskq.qsize() >= self.taskq_size_limit:
                            time.sleep(self._taskgen_sleep_time)
                        else:
                            self.submit_task(item)
                            item = None
        except (KeyboardInterrupt, Exception) as ex:
            self.fatalq.put((sys.exc_info(), None))

    def is_result_ok(self, req, res):
        if res.error:
            return False
        elif isinstance(req, CallbackRequest):
            return True
        elif (
                0 < res.status < 400
                or res.status == 404
                or (
                    req.config['extra_valid_status']
                    and res.status in req.config['extra_valid_status']
                )
            ):
            return True
        else:
            return False

    def thread_result_processor(self, pause):
        try:
            error_ctx = None
            while not self.shutdown_event.is_set():
                if pause.pause_event.is_set():
                    pause.process_pause()
                try:
                    result = self.resultq.get(True, 0.1)
                except Empty:
                    pass
                else:
                    self.stat.inc('crawler:request-processed')
                    try:
                        if (
                                result['request']['raw']
                                or self.is_result_ok(
                                    result['request'],
                                    result['response'],
                                )
                            ):
                            self.process_ok_result(result)
                        else:
                            self.process_fail_result(result)
                    except Exception as ex:
                        error_ctx = collect_error_context(result['request'])
                        raise
        except (KeyboardInterrupt, Exception) as ex:
            self.fatalq.put((sys.exc_info(), error_ctx))

    def thread_fatalq_processor(self):
        try:
            while not self.shutdown_event.is_set():
                try:
                    exc_info, ctx = self.fatalq.get(True, 0.1)
                except Empty:
                    pass
                else:
                    self.shutdown_event.set()
                    self.fatal_error_happened.set()
                    logging.error('Fatal exception')
                    if ctx:
                        logging.error('Exception context:')
                        for key, val in sorted(ctx.items(), key=lambda x: x[0]):
                            logging.error(' * %s: %s' % (key, val))
                    else:
                        logging.error('Exception context: N/A')
                    logging.error(''.join(format_exception(*exc_info)))
                    if not isinstance(exc_info[1], KeyboardInterrupt):
                        self.log_error(exc_info, ctx)
        except Exception as ex:
            self.shutdown_event.set()
            raise

    def log_error(self, exc_info, ctx=None):
        ctx = ctx or {}
        ctx.update({
            'crawler_id': self.__class__.__name__,
            'date': datetime.utcnow().isoformat(),
        })
        self.error_logger.log_error(exc_info, ctx)

    def thread_manager(self, th_task_gen, pauses):
        try:
            th_task_gen.join()

            def system_is_busy():
                return (
                    self.taskq.qsize()
                    or self.resultq.qsize()
                    or len(self.network.active_handlers)
                )

            while not self.shutdown_event.is_set():
                while system_is_busy():
                    if self.shutdown_event.is_set():
                        return
                    time.sleep(0.1)

                for pause in pauses:
                    pause.pause_event.set()
                ok = True
                for pause in pauses:
                    if not pause.is_paused.wait(0.1):
                        ok = False
                        break
                if not ok:
                    for pause in pauses:
                        pause.pause_event.clear()
                        pause.resume_event.set()
                else:
                    if not system_is_busy():
                        for pause in pauses:
                            pause.pause_event.clear()
                            pause.resume_event.set()
                        self.shutdown_event.set()
                        return
                    else:
                        for pause in pauses:
                            pause.pause_event.clear()
                            pause.resume_event.set()
        except (KeyboardInterrupt, Exception) as ex:
            self.fatalq.put((sys.exc_info(), None))

    def process_ok_result(self, result):
        self.stat.inc('crawler:request-ok')
        name = result['request'].config['name']
        handler = getattr(self, 'handler_%s' % name)
        try:
            handler_result = handler(
                result['request'],
                result['response'],
            )
            try:
                iter(handler_result)
            except TypeError:
                pass
            else:
                for item in handler_result:
                    if isinstance(item, BaseRequest):
                        self.submit_task(item)
                    else:
                        raise Exception(
                            'Handler yielded non request task: %s' % item
                        )
        except Exception as ex:
            self.stat.inc('result-handler-error:%s' % get_error_tag(ex))
            if self.stop_on_handler_error:
                raise
            else:
                logging.exception('Error in result handler')
                ctx = collect_error_context(result['request'])
                self.log_error(sys.exc_info(), ctx)

    def process_fail_result(self, result):
        self.stat.inc('crawler:request-fail')
        req = result['request']

        if result['response'].error:
            self.stat.inc('network-error:%s' % get_error_tag(
                result['response'].error
            ))
        if result['response'].status:
            self.stat.inc('http:status-%s' % result['response'].status)

        if req.retry_count + 1 <= self.retry_limit:
            self.stat.inc('crawler:request-retry')
            req.retry_count += 1
            req.priority = req.priority - 1
            self.submit_task(req)
        else:
            self.stat.inc('crawler:request-rejected')
            name = result['request'].config['name']
            handler = getattr(self, 'rejected_%s' % name, None)
            if handler is None:
                handler = self.default_rejected_handler
            try:
                handler(result['request'], result['response'])
            except Exception as ex:
                self.stat.inc('rejected-handler-error:%s' % get_error_tag(ex))
                if self.stop_on_handler_error:
                    raise
                else:
                    logging.exception('Error in rejected result handler')
                    ctx = collect_error_context(req)
                    self.log_error(sys.exc_info(), ctx)

    def default_rejected_handler(self, req, res):
        pass

    def run_hook(self):
        pass

    def thread_debug(self):
        try:
            while not self.shutdown_event.is_set():
                stat = []
                now = time.time()
                for hdl in self.network.active_handlers:
                    stat.append((hdl, self.network.registry[hdl]['start']))
                with open('var/crawler.debug', 'w') as out:
                    for hdl, start in list(sorted(stat, key=lambda x: (x[1] or now), reverse=True)):
                        req = self.network.registry[hdl]['request']
                        out.write('%.2f - [#%s] - %s\n' % (
                            (now - start) if start else 0,
                            req.retry_count if req else 'NA',
                            urlsplit(req['url']).netloc if req else 'NA',
                        ))
                    out.write('Active handlers: %d\n' % len(self.network.active_handlers))
                    out.write('Idle handlers: %d\n' % len(self.network.idle_handlers))
                    out.write('Taskq size: %d\n' % self.taskq.qsize())
                    out.write('Resultq size: %d\n' % self.resultq.qsize())

                total = 0
                count = 0
                for hdl, start in stat:
                    if start:
                        total += (now - start)
                        count += 1
                logging.debug('Median handler time: %.2f' % ((total / count) if count else 0))

                self.shutdown_event.wait(3)
        except (KeyboardInterrupt, Exception) as ex:
            self.fatalq.put((sys.exc_info(), None))


    def thread_network(self):
        try:
            self.network.run()
        except (KeyboardInterrupt, Exception) as ex:
            self.fatalq.put((sys.exc_info(), None))

    def shutdown(self):
        for name in self.dataopq.keys():
            self.enq_dataop(name, None, force_dump=True)
        self.shutdown_hook()

    def shutdown_hook(self):
        pass

    def run(self):
        try:
            self.run_hook()
            self._run_started = time.time()

            th_fatalq_proc = Thread(target=self.thread_fatalq_processor)
            th_fatalq_proc.start()

            th_task_gen = Thread(target=self.thread_task_generator)
            th_task_gen.start()

            th_debug = Thread(target=self.thread_debug)
            if self.debug:
                th_debug.start()

            pauses = [self.network_pause]
            result_workers = []
            for _ in range(self.result_workers):
                pause = Pause()
                th = Thread(
                    target=self.thread_result_processor,
                    args=[pause],
                ) 
                pauses.append(pause)
                th.start()
                result_workers.append(th)

            th_manager = Thread(
                target=self.thread_manager,
                args=[th_task_gen, pauses],
            )
            th_manager.start()

            th_network = Thread(
                target=self.thread_network,
            )
            th_network.start()

            th_manager.join()
            th_fatalq_proc.join()
            th_task_gen.join()
            if self.debug:
                th_debug.join()
            [x.join() for x in result_workers]

        except (Exception, KeyboardInterrupt):
            self.fatal_error_happened.set()
            raise
        finally:
            self.shutdown_event.set()
            self.shutdown()



class Data(object):
    __slots__ = (
        'name',
        'meta',
    )

    def __init__(self, name, meta=None):
        self.name = name
        self.meta = meta or {}

    def __getitem__(self, key):
        return self.meta[key]

    def as_data(self):
        return {
            'name': self.name,
            'meta': self.meta,
        }

    @classmethod
    def from_data(cls, data):
        req = Data()
        for key in ('name', 'storage',):
            setattr(req, key, data[key])
        return req



import logging
from collections import defaultdict
from threading import Event
import time

debug_logger = logging.getLogger('debug')
DEBUG_TIMES = defaultdict(int)


class Pause(object):
    def __init__(self):
        self.pause_event = Event()
        self.is_paused = Event()
        self.resume_event = Event()

    def process_pause(self):
        self.pause_event.clear()
        self.resume_event.clear() # as a precaution
        self.is_paused.set()
        self.resume_event.wait()
        self.is_paused.clear()
        self.resume_event.clear()


def debug(msg, *args):
    if time.time() - DEBUG_TIMES[msg] >= 0.5:
        debug_logger.debug(msg, *args)
        DEBUG_TIMES[msg] = time.time()



import pytest

from ioweb.request import CallbackRequest


def default_config():
    req = CallbackRequest()
    keys = [
        'name',
        'timeout',
        'raw',
        'network_callback',
    ]
    assert set(keys) == set(req.config.keys())



"""
No gevent monkey patching is here.
Monkey patching is done by `ioweb` console command
which is imported from `ioweb_gevent` package
in setup.py entry point
"""
__version__ = '0.0.11'

from .session import Session
from .request import Request, CallbackRequest
from .data import Data
from .response import Response
from .crawler import Crawler
from .transport import Urllib3Transport
from .task_generator import TaskGenerator
from .error import *



"""
TODO: remove curl references
"""
class IowebError(Exception):
    pass


class IowebConfigError(IowebError):
    pass


class NetworkError(IowebError):
    """
    NetworkError instance stores original transport-specific error
    in `transport_error` attribute.
    """
    def __init__(self, *args):
        self.errmsg = None
        self.transport_error = None
        if len(args) > 0:
            self.errmsg = args[0]
        if len(args) > 1:
            self.transport_error = args[1]


class ProxyError(NetworkError):
    """
    Error occured because of proxy server failed.
    """


class DataWriteError(NetworkError):
    """
    An error occurred when writing received data to a local file,
    or an error was returned to libcurl from a write callback.
    """


class OperationTimeoutError(NetworkError):
    """
    Operation timeout. The specified time-out period was reached
    according to the conditions.
    """


class ConnectError(NetworkError):
    """
    Failed to connect() to host or proxy.
    """


class AuthError(NetworkError):
    """
    The remote server denied client to login (Added in 7.13.1)
    """


class TooManyRedirectsError(NetworkError):
    """
    Too many redirects. When following redirects, libcurl hit the
    maximum amount. Set your limit with CURLOPT_MAXREDIRS.
    """


class ResolveHostError(NetworkError):
    """
    Couldn't resolve host. The given remote host was not resolved.
    """


class InvalidUrlError(NetworkError):
    """
    The URL was not properly formatted.
    """


class TooManyRedirects(NetworkError):
    """
    Number of redirects reached configured limit.
    """


class MalformedResponseError(NetworkError):
    """
    Raised when parser fails to parse response HTTP headers
    """


ERRNO_CLASS_MAPPING = {
    3: InvalidUrlError,
    6: ResolveHostError,
    7: ConnectError,
    23: DataWriteError,
    28: OperationTimeoutError,
    47: TooManyRedirectsError,
    67: AuthError,
}


def get_error_tag(err):
    original_err = err
    if isinstance(err, NetworkError) and err.transport_error:
        original_err = err.transport_error
    return original_err.__class__.__name__.lower()


def collect_error_context(req):
    ctx = {}
    if 'url' in req.config:
        ctx['req_url'] = req['url']
    ctx['req_name'] = req.config.get('name', None)
    if req.error_context:
        try:
            ctx.update(req.error_context(req))
        except Exception as ex:
            ctx.update({
                'internal_error': (
                    'error_context callback failed with %s'
                    % ex
                )
            })
    return ctx


class DataNotValid(IowebError):
    """Used to signal that expected data was not found
    i.e. it needs to retry network query.
    """



from io import BytesIO
from collections import defaultdict
#from base64 import b64encode
try:
    import ujson as json
except ImportError:
    import json

import defusedxml.lxml
from lxml.etree import HTMLParser
from selection import XpathSelector
from cssselect import HTMLTranslator

#from urllib3.contrib import pyopenssl


class CssSelector(XpathSelector):
    __slots__ = ()

    def process_query(self, query):
        xpath_query = HTMLTranslator().css_to_xpath(query)
        return super(CssSelector, self).process_query(xpath_query)


class Response(object):
    __slots__ = (
        '_bytes_body',
        '_cached',
        'cert',
        'status',
        'error',
        'headers',
        'meta',
        '_cached_dom',
    )

    def __init__(self):
        self._bytes_body = BytesIO()
        self.headers = None
        self.cert = None
        self.status = None
        self.error = None
        self.meta = {}
        self._cached_dom = None

    def write_bytes_body(self, data):
        return self._bytes_body.write(data)

    @property
    def data(self):
        return self.bytes_body

    @property
    def text(self):
        return self.data.decode('utf-8')

    @property
    def bytes_body(self):
        return self._bytes_body.getvalue()

    @property
    def json(self):
        return json.loads(self.text)

    def get_content_type(self):
        if self.headers:
            return (
                self.headers.get('content-type', '').split(';')[0].strip()
            )
        else:
            return None

    def dom(self):
        if self._cached_dom is None:
            res = defusedxml.lxml.parse(BytesIO(self.data), HTMLParser())
            self._cached_dom = res.getroot()
        return self._cached_dom

    def xpath(self, query):
        return XpathSelector(self.dom()).select(query)

    def css(self, query):
        return CssSelector(self.dom()).select(query)



from pprint import pprint
import sys
import re
import time
import os.path
import json
import logging
from argparse import ArgumentParser
from importlib import import_module

from .crawler import Crawler

logger = logging.getLogger('crawler.cli')


def find_crawlers_in_module(mod, reg):
    for key in dir(mod):
        val = getattr(mod, key)
        if (
                isinstance(val, type)
                and issubclass(val, Crawler)
                and val is not Crawler
            ):
            logger.error(
                'Found crawler %s in module %s',
                val.__name__, mod.__file__
            )
            reg[val.__name__] = val


def collect_crawlers():
    reg = {}

    # Give crawlers in current directory max priority
    # Otherwise `/web/crawler/crawlers` packages are imported
    # when crawler is installed with `pip -e /web/crawler`
    sys.path.insert(0, os.getcwd())

    for location in ('crawlers',):
        try:
            mod = import_module(location)
        except ImportError as ex:
            #if path not in str(ex):
            logger.exception('Failed to import %s', location)
        else:
            if getattr(mod, '__file__', '').endswith('__init__.py'):
                dir_ = os.path.split(mod.__file__)[0]
                for fname in os.listdir(dir_):
                    if (
                        fname.endswith('.py')
                        and not fname.endswith('__init__.py')
                    ):
                        target_mod = '%s.%s' % (location, fname[:-3])
                        try:
                            mod = import_module(target_mod)
                        except ImportError as ex:
                            #if path not in str(ex):
                            logger.exception('Failed to import %s', target_mod)
                        else:
                            find_crawlers_in_module(mod, reg)
            else:
                find_crawlers_in_module(mod, reg)

    return reg


def setup_logging(network_logs=False):#, control_logs=False):
    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger('urllib3.connectionpool').setLevel(level=logging.ERROR)
    logging.getLogger('urllib3.util.retry').setLevel(level=logging.ERROR)
    logging.getLogger('urllib3.poolmanager').setLevel(level=logging.ERROR)
    logging.getLogger('ioweb.urllib3_custom').setLevel(level=logging.ERROR)
    if not network_logs:
        logging.getLogger('ioweb.network_service').propagate = False
    #if not control_logs:
    #    logging.getLogger('crawler.control').propagate = False


def format_elapsed_time(total_sec):
    hours = minutes = 0
    if total_sec > 3600:
        hours, total_sec = divmod(total_sec, 3600)
    if total_sec > 60:
        minutes, total_sec = divmod(total_sec, 60)
    return '%02d:%02d:%.2f' % (hours, minutes, total_sec)


def get_crawler(crawler_id):
    reg = collect_crawlers()
    if crawler_id not in reg:
        sys.stderr.write(
            'Could not find %s crawler\n' % crawler_id
        )
        sys.exit(1)
    else:
        return reg[crawler_id]


def run_subcommand_crawl(opts):
    setup_logging(network_logs=opts.network_logs)#, control_logs=opts.control_logs)
    cls = get_crawler(opts.crawler_id)
    extra_data = {}
    for key in cls.extra_cli_args():
        opt_key = 'extra_%s' % key.replace('-', '_')
        extra_data[key] = getattr(opts, opt_key)
    bot = cls(
        network_threads=opts.network_threads,
        extra_data=extra_data,
        debug=opts.debug,
    )
    try:
        if opts.profile:
            import cProfile
            import pyprof2calltree
            import pstats

            profile_file = 'var/%s.prof' % opts.crawler_id
            profile_tree_file = 'var/%s.prof.out' % opts.crawler_id

            prof = cProfile.Profile()
            try:
                prof.runctx('bot.run()', globals(), locals())
            finally:
                stats = pstats.Stats(prof)
                stats.strip_dirs()
                pyprof2calltree.convert(stats, profile_tree_file)
        else:
            bot.run()
    except KeyboardInterrupt:
        bot.fatal_error_happened.set()
    print('Stats:')
    for key, val in sorted(bot.stat.total_counters.items()):
        print(' * %s: %s' % (key, val))
    if bot._run_started:
        print('Elapsed: %s' % format_elapsed_time(time.time() - bot._run_started))
    else:
        print('Elapsed: NA')
    if bot.fatal_error_happened.is_set():
        sys.exit(1)
    else:
        sys.exit(0)


def run_subcommand_foo(parser, opts):
    print('COMMAND FOO')


def command_ioweb():
    parser = ArgumentParser()#add_help=False)

    crawler_cls = None
    if len(sys.argv) > 2:
        if sys.argv[1] == 'crawl':
            crawler_cls = get_crawler(sys.argv[2])

    subparsers = parser.add_subparsers(
        dest='command',
        title='Subcommands of ioweb command',
        description='',
    )

    # Crawl
    crawl_subparser = subparsers.add_parser(
        'crawl', help='Run crawler',
    )
    crawl_subparser.add_argument('crawler_id')
    crawl_subparser.add_argument('-t', '--network-threads', type=int, default=1)
    crawl_subparser.add_argument('-n', '--network-logs', action='store_true', default=False)
    crawl_subparser.add_argument('-p', '--profile', action='store_true', default=False)
    crawl_subparser.add_argument('--debug', action='store_true', default=False)
    #parser.add_argument('--control-logs', action='store_true', default=False)
    if crawler_cls:
        crawler_cls.update_arg_parser(crawl_subparser)

    # Foo
    foo_subparser = subparsers.add_parser(
        'foo', help='Just test subcommand',
    )

    opts = parser.parse_args()
    if opts.command == 'crawl':
        run_subcommand_crawl(opts)
    else:
        if opts.command == 'foo':
            run_subcommand_foo(opts)
        else:
            parser.print_help()



from itertools import cycle
import re
from urllib.request import urlopen
from random import choice
import logging

RE_PROXYLINE = re.compile(r'^([^:]+):(\d+)$')
RE_PROXYLINE_AUTH = re.compile(
    r'^([^:]+):(\d+):(.+?):(.+)$'
)
logger = logging.getLogger('crawler.proxylist')


class Proxy(object):
    __slots__ = ('host', 'port', 'user', 'password', 'proxy_type')

    def __init__(
            self, host=None, port=None, user=None,
            password=None, proxy_type=None
        ):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.proxy_type = proxy_type

    def address(self):
        return '%s:%s' % (self.host, self.port)

    def auth(self):
        if self.user:
            return '%s:%s' % (self.user, self.password)
        else:
            return None


class ProxyList(object):
    def __init__(self, proxy_type='http'):
        self._servers = []
        self._source = None
        self.proxy_type = proxy_type

    @classmethod
    def create_from_source(cls, src_type, src, **kwargs):
        pl = ProxyList(**kwargs)
        pl.load_source(src_type, src)
        return pl

    def load_source(self, source_type, source):
        assert source_type in ('file', 'url', 'list'), \
            'Incorrect source type: %s' % source_type
        if source_type == 'file':
            self.load_file(source)
        elif source_type == 'url':
            self.load_url(source)
        elif source_type == 'list':
            self.load_list(source)

    def load_file(self, path):
        self._source = {
            'type': 'file',
            'location': path,
        }
        with open(path) as inp:
            self.load_from_rawdata(
                inp.read().splitlines()
            )
            
    def load_url(self, url):
        self._source = {
            'type': 'url',
            'location': url,
        }
        return self.load_from_rawdata(
            urlopen(url).read().decode('utf-8').splitlines()
        )

    def load_list(self, items):
        self._source = {
            'type': 'list',
            'location': None,
        }
        return self.load_from_rawdata(items)

    def load_from_rawdata(self, lines): 
        servers = []
        for line in lines:
            line = line.strip()
            match = RE_PROXYLINE.match(line)
            if not match:
                match = RE_PROXYLINE_AUTH.match(line)
                if not match:
                    logger.error('Invalid proxy line: %s' % line)
                else:
                    host, port, user, password = match.groups()
            else:
                host, port = match.groups()
                user, password = None, None
            port = int(port)
            servers.append(
                Proxy(
                    host=host, port=port,
                    user=user, password=password,
                    proxy_type=self.proxy_type,
                )
            )
        if servers:
            self._servers = servers
            self._servers_iter = cycle(self._servers)

    def random_server(self):
        return choice(self._servers)

    def next_server(self):
        return next(self._servers_iter)

    def reload(self):
        if self._source['type'] == 'list':
            pass
        else:
            self.load_source(
                self._source['type'],
                self._source['location'],
            )



import pytest

from ioweb.request import Request


def default_config():
    req = Request()
    keys = [
        'method',
        'name',
        'url',
        'max_redirects',
        'follow_redirect',
        'timeout',
        'connect_timeout',
        'resolve',
        'raw',
        'headers',
        'content_encoding',
        'decode_content',
        'content_read_limit',
        'priority',
        'extra_valid_status',
        'proxy',
        'proxy_auth',
        'proxy_type',
        'data',
        'multipart',
        'verify',
    ]
    assert set(keys) == set(req.config.keys())



from setuptools import setup, find_packages
import os

ROOT = os.path.dirname(os.path.realpath(__file__))

with open('README.md') as inp:
    readme_content = inp.read()

setup(
    name = 'ioweb',
    version = '0.0.11',
    author = 'Gregory Petukhov',
    author_email = 'lorien@lorien.name',
    maintainer='Gregory Petukhov',
    maintainer_email='lorien@lorien.name',
    url='https://github.com/lorien/ioweb',
    description = 'Web Scraping Framework',
    long_description = readme_content,
    long_description_content_type='text/markdown',
    packages = find_packages(exclude=['test']),
    download_url='https://github.com/lorien/ioweb/releases',
    license = "MIT",
    entry_points = {
        'console_scripts': [
            'ioweb=ioweb_gevent.cli:command_ioweb',
        ],
    },
    install_requires = [
        'urllib3==1.25.6',
        'pyopenssl',
        'cryptography',
        'idna',
        'certifi',
        'cachetools',
        'gevent',
        'pysocks',
        'lxml',
        'defusedxml',
        'selection',
        'cssselect',
    ],
    keywords='web scraping network crawling cralwer spider pycurl',
    classifiers = [
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



from .stat import Stat


class TaskGenerator(object):
    def __init__(self):
        self.stat = Stat()
        self.init_hook()

    def init_hook(self):
        pass

    def task_generator(self):
        raise NotImplementedError



import pytest

from ioweb.request import BaseRequest


def test_default_meta():
    req = BaseRequest()
    assert req.meta == {}


def test_custom_default_config():
    class CustomRequest(BaseRequest):
        def get_default_config(self):
            return {
                'foo': 'bar',
            }

    req = CustomRequest()
    assert req.config['foo'] == 'bar'


def test_setup_invalid_key():
    class CustomRequest(BaseRequest):
        def get_default_config(self):
            return {
                'foo': 'bar',
            }

    req = CustomRequest()
    with pytest.raises(AssertionError):
        req.setup(zzz=1)


def test_getitem_method():
    class CustomRequest(BaseRequest):
        def get_default_config(self):
            return {
                'foo': 'bar',
            }

    req = CustomRequest()
    assert req['foo'] == 'bar'


def test_getitem_method_invalid_key():
    class CustomRequest(BaseRequest):
        def get_default_config(self):
            return {
                'foo': 'bar',
            }

    req = CustomRequest()
    with pytest.raises(KeyError):
        assert req['zzz']


def test_as_data():
    class CustomRequest(BaseRequest):
        def get_default_config(self):
            return {
                'foo': 'bar',
            }

    req = CustomRequest()
    assert req.as_data()['config'] == {'foo': 'bar'}
    assert all(
        x in req.as_data()
        for x in ('config', 'meta', 'priority', 'retry_count')
    )


def test_lt():
    req1 = BaseRequest(priority=1)
    req2 = BaseRequest(priority=2)
    req3 = BaseRequest(priority=3)
    assert req2 > req1
    assert req1 < req3


def test_eq():
    req1 = BaseRequest(priority=1)
    req2 = BaseRequest(priority=2)
    req3 = BaseRequest(priority=2)
    assert req1 != req2
    assert req2 == req3


def test_from_data():
    data = {
        'config': {'foo': 'bar'},
        'meta': {'name': 'Putin'},
        'priority': 1,
        'retry_count': 0,
    }
    req = BaseRequest.from_data(data)
    assert req.config['foo'] == 'bar'
    assert req.meta['name'] == 'Putin'



"""
Credit: https://github.com/Roadmaster/forcediphttpsadapter/blob/master/forcediphttpsadapter/adapters.py
"""
import logging

from urllib3.util import connection
from urllib3.connectionpool import HTTPSConnectionPool, HTTPConnectionPool
from urllib3.connection import HTTPSConnection, DummyConnection, HTTPConnection
from urllib3 import PoolManager
from urllib3.poolmanager import SSL_KEYWORDS
from socket import error as SocketError, timeout as SocketTimeout
from urllib3.exceptions import (
    ConnectTimeoutError,
    NewConnectionError,
)
from cachetools import TTLCache

log = logging.getLogger(__name__)


class CustomHttpConnection(HTTPConnection, object):
    def __init__(self, *args, **kwargs):
        self._custom_ip = kwargs.pop('custom_ip', None)
        super(CustomHttpConnection, self).__init__(*args, **kwargs)

    @property
    def _dns_host(self):
        return self._custom_ip if self._custom_ip else self._dns_host_value

    @_dns_host.setter
    def _dns_host(self, value):
        self._dns_host_value = value


class CustomHttpsConnection(HTTPSConnection, object):
    def __init__(self, *args, **kwargs):
        self._custom_ip = kwargs.pop('custom_ip', None)
        super(CustomHttpsConnection, self).__init__(*args, **kwargs)

    @property
    def _dns_host(self):
        return self._custom_ip if self._custom_ip else self._dns_host_value

    @_dns_host.setter
    def _dns_host(self, value):
        self._dns_host_value = value


class CustomHttpConnectionPool(HTTPConnectionPool):
    ConnectionCls = CustomHttpConnection

    def __init__(self, *args, **kwargs):
        self._custom_ip = kwargs.pop('custom_ip', None)
        super(CustomHttpConnectionPool, self).__init__(*args, **kwargs)
        self.conn_kw['custom_ip'] = self._custom_ip


class CustomHttpsConnectionPool(HTTPSConnectionPool):
    ConnectionCls = CustomHttpsConnection

    def __init__(self, *args, **kwargs):
        self._custom_ip = kwargs.pop('custom_ip', None)
        super(CustomHttpsConnectionPool, self).__init__(*args, **kwargs)
        self.conn_kw['custom_ip'] = self._custom_ip

    def __str__(self):
        return ('%s(host=%r, port=%r, custom_ip=%s)' % (
            type(self).__name__,
            self.host,
            self.port,
            self._custom_ip,
        ))


class CustomPoolManager(PoolManager):
    def __init__(self, *args, **kwargs):
        super(CustomPoolManager, self).__init__(*args, **kwargs)
        self.resolving_cache = TTLCache(maxsize=10000, ttl=(60 * 60))
        self.pool_classes_by_scheme = {
            'http': CustomHttpConnectionPool,
            'https': CustomHttpsConnectionPool,
        }

    def _new_pool(self, scheme, host, port, request_context=None):
        if request_context is None:
            request_context = self.connection_pool_kw.copy()
        custom_ip = self.resolving_cache.get(host, None)
        if custom_ip:
            request_context['custom_ip'] = custom_ip
        return super(CustomPoolManager, self)._new_pool(
            scheme, host, port, request_context
        )



from pprint import pprint, pformat
from collections import defaultdict, deque
import time
import logging
from importlib import import_module
from threading import Thread
from copy import deepcopy
import sys
from datetime import datetime

logger = logging.getLogger('ioweb.stat')

class Stat(object):
    default_key_aliases = {
        'crawler:request-processed': 'req',
        'crawler:request-proxy-processed': 'req-proxy',
        'crawler:request-ok': 'req-ok',
        'crawler:request-retry': 'req-retry',
        'crawler:request-fail': 'req-fail',
        'crawler:request-rejected': 'req-rejected',
    }
    ignore_prefixes = (
        'http:',
        'network-error:',
    )
    def __init__(
            self,
            # logging
            speed_keys=None,
            logging_enabled=True,
            logging_interval=3,
            key_aliases=None,
            # export
            #shard_interval = 10,
            export=None,
            export_interval=5,
            # fatalq
            fatalq=None,
        ):
        # Arg: speed_keys
        if speed_keys is None:
            speed_keys = []
        elif isinstance(speed_keys, str):
            speed_keys = [speed_keys]
        self.speed_keys = speed_keys
        # Arg: logging_enabled
        self.logging_enabled = logging_enabled
        # Arg: logging_interval
        self.logging_interval = logging_interval
        # Arg: key_aliases
        self.key_aliases = deepcopy(self.default_key_aliases)
        if key_aliases:
            self.key_aliases.update(key_aliases)

        # Arg: fatalq
        self.fatalq = fatalq

        # Arg: shard_interval
        #self.shard_interval = shard_interval

        # Logging
        self.total_counters = defaultdict(int)
        self.moment_counters = {}
        self.logging_time = 0
        if self.logging_enabled:
            self.th_logging = Thread(target=self.thread_logging)
            self.th_logging.daemon = True
            self.th_logging.start()

        # Setup exporting in last case
        # Arg: export
        self.th_export = None
        self.export_config = export
        self.export_driver = None
        if self.export_config:
            self.setup_export_driver(self.export_config)

        # Args: export_interval
        self.export_interval = export_interval

        # Export
        #self.shard_counters = {}
        if self.export_driver:
            self.start_export_thread()

    def start_export_thread(self):
        if self.export_driver and not self.th_export:
            self.th_export = Thread(target=self.thread_export)
            self.th_export.daemon = True
            self.th_export.start()

        # Internal
        self.service_time = 0
        self.service_interval = 1

    def setup_export_driver(self, cfg):
        self.export_config = cfg
        mod_path, cls_name = cfg['driver'].split(':', 1)
        driver_mod = import_module(mod_path)
        driver_cls = getattr(driver_mod, cls_name)
        self.export_driver = driver_cls(
            tags=cfg.get('tags', {}),
            connect_options=cfg.get('connect_options', {})
        )
        if self.th_export:
            raise Exception('Stat export thread already created')
        else:
            self.start_export_thread()

    def build_eps_string(self, now):
        now_int = int(now)
        eps = defaultdict(int)
        interval = 30
        for ts in range(now_int - interval, now_int):
            for key in self.speed_keys:
                try:
                    eps[key] += self.moment_counters[ts][key]
                except KeyError:
                    eps[key] += 0

        ret = []
        for key, val in eps.items():
            label = self.key_aliases.get(key, key)
            val_str = '%.1f' % (val / interval)
            if val_str == '0.0' and val > 0:
                val_str = '0.0+'
            ret.append('%s: %s' % (label, val_str))
        return ', '.join(ret)

    def build_counter_string(self):
        ret = []
        for key in sorted(list(self.total_counters.keys())):
            if not key.startswith(self.ignore_prefixes):
                label = self.key_aliases.get(key, key)
                val = self.total_counters[key]
                ret.append('%s: %d' % (label, val))
        return ', '.join(ret)

    def render_moment(self, now=None):
        if now is None:
            now = time.time()
        eps_str = self.build_eps_string(now)
        counter_str = self.build_counter_string()
        return 'EPS: %s | TOTAL: %s' % (eps_str, counter_str)

    def thread_logging(self):
        try:
            while True:
                now = time.time()
                logger.debug(self.render_moment(now))
                # Sleep `self.logging_interval` seconds minus time spent on logging
                sleep_time = (
                    self.logging_interval + (time.time() - now)
                )
                time.sleep(sleep_time)
        except (KeyboardInterrupt, Exception) as ex:
            if self.fatalq:
                self.fatalq.put((sys.exc_info(), None))
            else:
                raise

    def thread_export(self):
        try:
            prev_counters = None
            while True:
                now = time.time()
                counters = deepcopy(self.total_counters)
                if prev_counters:
                    delta_counters = dict(
                        (x, counters[x] - prev_counters.get(x, 0))
                        for x in counters.keys()
                    )
                else:
                    delta_counters = counters
                prev_counters = counters
                self.export_driver.write_events(delta_counters)
                sleep_time = (
                    self.export_interval + (time.time() - now)
                )
                time.sleep(sleep_time)
        except (KeyboardInterrupt, Exception) as ex:
            if self.fatalq:
                self.fatalq.put((sys.exc_info(), None))
            else:
                raise

    def inc(self, key, count=1):
        now_int = int(time.time())
        #shard_ts = now_int - now_int % self.shard_interval
        #shard_slot = self.shard_counters.setdefault(shard_ts, defaultdict(int))
        moment_slot = self.moment_counters.setdefault(now_int, defaultdict(int))

        moment_slot[key] += count
        #shard_slot[key] += count
        self.total_counters[key] += count


class InfluxdbExportDriver(object):
    def __init__(self, connect_options, tags):
        self.connect_options = deepcopy(connect_options)
        self.client = None
        self.tags = deepcopy(tags)
        self.database_created = False
        self.connect()

    def connect(self):
        from influxdb import InfluxDBClient

        self.client = InfluxDBClient(**self.connect_options)

    def write_events(self, snapshot):
        from requests import RequestException

        if not self.database_created:
            self.client.create_database(self.connect_options['database'])
            self.database_created = True
        if snapshot:
            data = {
                "measurement": "crawler_counter",
                "tags": self.tags,
                "time": datetime.utcnow().isoformat(),
                "fields": dict((
                    (x, y) for x, y in snapshot.items()
                )),
            }
            while True:
                try:
                    self.client.write_points([data])
                except RequestException:
                    logger.exception('Fail to send metrics')
                    time.sleep(1)
                    # reconnecting
                    while True:
                        try:
                            self.connect()
                        except RequestException:
                            logger.exception(
                                'Fail to reconnect to metric database'
                            )
                            time.sleep(1)
                        else:
                            break
                else:
                    break
